/*******************************************************************************
 *	BdsTestValidateData.cc	Validate test data and meta data on BdsServer
 *				T.Barnaby,	BEAM Ltd,	2009-11-23
 *******************************************************************************
 *
 * This program validates the test data and MetaData in the BDS system.
 */
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <BdsD.h>
#include <BdsC.h>
#include <BFile.h>
#include <BdsLib.h>
#include <sys/stat.h>

using namespace std;
using namespace Bds;

class Config {
public:
				Config();
	int			verbose;
	int			sampleMultiplexed;
	Selection		selection;
	BString			fileName;
};

Config::Config(){
	verbose = 0;
	sampleMultiplexed = 0;
}

BError waterMarkRead(DataBlock& dataBlock, BUInt channel, BDictString& info){
	BError		err;
	BString		mark;
	double*		d = &dataBlock.channelData[channel][0];
	int		n;
	char		ch;
	BList<BString>	list;
	
	for(n = 0; n < 1024; n++){
//		printf("Data: %llu\n", (BInt64)d[n]);
		ch = ((BInt64)d[n]) & (0xFFULL);
		if(ch == 0)
			break;
		mark += ch;
	}
	
	if(n >= 1024)
		return err.set(1, BString("No water mark found in dataBlock: ") + dataBlock.startTime.getString());

	list = mark.getTokenList(',');
	
//	printf("Mark: %s\n", mark.retStr());
	if(list.size() != 9)
		return err.set(1, BString("No valid water mark in dataBlock: ") + dataBlock.startTime.getString() + " list size: " + list.size());

	info["network"] = list[0];
	info["array"] = list[1];
	info["startTime"] = list[2];
	info["endTime"] = list[3];
	info["station"] = list[4];
	info["channel"] = list[5];
	info["source"] = list[6];
	info["blockSize"] = list[7];
	info["checksum"] = list[8];
	
	return err;
}

BError testValidateDataBlock(Config& config, BUInt channel, BUInt blockNumber, DataBlock& data){
	BError			err;
	BUInt			c;
	BUInt			s;
	BIter			i;
	double			checkSum;
	BDictString		info;

	for(c = 0; c < data.channelData.size(); c++){
		checkSum = 0;
		for(s = 0; s < data.channelData[c].size(); s++){
			checkSum += double((BInt64)data.channelData[c][s] & ~(0xFFULL));
		}

		// Test water mark
		if(err = waterMarkRead(data, c, info))
			return err;

		if(config.verbose > 1){
			printf("Channel: %u Block: %u\n", channel, blockNumber);
			for(info.start(i); !info.isEnd(i); info.next(i)){
				printf("\t%s: %s\n", info.key(i).retStr(), info[i].retStr());
			}
		}

		// Check data
		if(info["network"] != config.selection.channels[channel + c].network)
			return err.set(1, BString("Network incorrect: ") + info["network"] + " should be " + config.selection.channels[channel + c].network);
		if(info["station"] != config.selection.channels[channel + c].station)
			return err.set(1, BString("Station incorrect: ") + info["station"] + " should be " + config.selection.channels[channel + c].station);
		if(info["channel"] != config.selection.channels[channel + c].channel)
			return err.set(1, BString("Channel incorrect: ") + info["channel"] + " should be " + config.selection.channels[channel + c].channel);
		if(info["source"] != config.selection.channels[channel + c].source)
			return err.set(1, BString("Source incorrect: ") + info["source"] + " should be " + config.selection.channels[channel + c].source);
		if(info["checksum"] != BString::convert(checkSum, 1))
			return err.set(1, BString("CheckSum incorrect: ") + BString::convert(checkSum, 1) + " should be " + info["checksum"]);
	}
	
	return err;
}

BError testValidateData(DataAccess& bds, Config& config){
	BError			err;
	DataInfo		dataInfo;
	DataHandle		dataHandle;
	DataBlock		data;
	BUInt			c;
	BUInt			s;
	BUInt			b = 0;
	BIter			i;
	BTimeStamp		lastEndTime;

	if(config.verbose){
		printf("Validate Data\n");

		for(c = 0; c < config.selection.channels.number(); c++){
			printf("Chan: %d %s\n", c, bdsSelectionChannelInfo(config.selection, c).retStr());
		}
	}

	if(err = bds.dataSearch(config.selection, dataInfo))
		return err;
	
	if(!dataInfo.channels.size()){
		return err.set(1, "No Channels selected: Requires Network, Station/Array and Channels to be set to channels that exist");
	}
	
	if(err = bds.dataOpen(dataInfo, "r", "API", 0, dataHandle)){
		return err;
	}

	if(err = bds.dataGetInfo(dataHandle, 0, dataInfo)){
		return err;
	}

	if(config.sampleMultiplexed){
		while(1){
//			printf("Loop: Block: %u\n", blockNumber);
			if(err = bds.dataGetBlock(dataHandle, 0, 0, b, data)){
				if(err.getErrorNo() == ErrorEndOfData)
					err.set(0, "");
				return err;
			}

			if(data.endTime >= dataInfo.endTime)
				break;

			if(lastEndTime.isSet() && (lastEndTime != data.startTime))
				return err.set(1, BString("Block timestamp error: expecting: ") + lastEndTime.getString().retStr() + " got: " + data.startTime.getString().retStr());

			lastEndTime = data.endTime;

			if(err = testValidateDataBlock(config, 0, b, data))
				return err;

			b++;
		}
	}
	else {
		for(c = 0; c < dataInfo.channels.size(); c++){
			for(s = 0; s < dataInfo.channels[c].size(); s++){
				if(config.verbose)
					printf("Channel: %u Segment: %u NumBlocks: %u\n", c + 1, s + 1, dataInfo.channels[c][s].numBlocks);

				lastEndTime.clear();
				for(b = 0; b < dataInfo.channels[c][s].numBlocks; b++){
//					printf("Loop: Block: %u\n", b);
					if(err = bds.dataGetBlock(dataHandle, c + 1, s + 1, b, data)){
						return err;
					}

					if(lastEndTime.isSet() && (lastEndTime != data.startTime))
						return err.set(1, BString("Block timestamp error: expecting: ") + lastEndTime.getString().retStr() + " got: " + data.startTime.getString().retStr());

					lastEndTime = data.endTime;

					if(err = testValidateDataBlock(config, c, b, data))
						return err;
				}
			}
		}
	}
	bds.dataClose(dataHandle);

	return err;
}



void usage(void) {
	cerr << "Usage:\bdsTestAuto [options]\n";
	cerr << " -help              - Help on command line parameters\n";
	cerr << " -host <hostname>   - BDS Server host name\n";
	cerr << " -user <user:password> - user id and password\n";
	cerr << " -verbose           - verbose\n";
	cerr << " -startTime         - The StartTime\n";
	cerr << " -endTime           - The End Time\n";
	cerr << " -select <network:station:channel:source>	- Select a data set. Can have multiple -select's\n";
	cerr << " -sampleMultiplexed - Access the data in a sample multiplexed form\n";
}

static struct option options[] = {
		{ "?",			0, NULL, 0 },
		{ "h",			0, NULL, 0 },
		{ "help",		0, NULL, 0 },
		{ "host",		1, NULL, 0 },
		{ "user",		1, NULL, 0 },
		{ "verbose",		0, NULL, 0 },
		{ "startTime",		1, NULL, 0 },
		{ "endTime",		1, NULL, 0 },
		{ "select",		1, NULL, 0 },
		{ "sampleMultiplexed",	0, NULL, 0 },
		{ 0,0,0,0 }
};

int main(int argc, char** argv){
	BError			err;
	int			optIndex = 0;
	int			c;
	BString			s;
	DataAccess		bds;
	BString			hostName;
	BString			user = "test";
	BString			password = "beam00";
	Config			config;
	BString			s1,s2,s3,s4;

	// Set stdout to line buffering
	setlinebuf(stdout);

	hostName = getenv("BDS_HOST");
	if(hostName == "")
		hostName = "localhost";

	while((c = getopt_long_only(argc, argv, "", options, &optIndex)) == 0){
		s = options[optIndex].name;
		if(s == "help" || s == "h" || s == "?"){
			usage();
			return 1;
		}
		else if(s == "host"){
			hostName = optarg;
		}
		else if(s == "user"){
			s = optarg;
			user = s.pullToken(":");
			password = s;
		}
		else if(s == "verbose"){
			config.verbose++;
		}
		else if(s == "sampleMultiplexed"){
			config.sampleMultiplexed = 1;
		}
		else if(s == "startTime"){
			if(err = config.selection.startTime.setString(optarg)){
				cerr << "Error: StartTime format error\n";
				usage();
				return 1;
			}
		}
		else if(s == "endTime"){
			if(err = config.selection.endTime.setString(optarg)){
				cerr << "Error: EndTime format error\n";
				usage();
				return 1;
			}
		}
		else if(s == "select"){
			s = optarg;
			s1 = s.pullToken(":");
			s2 = s.pullToken(":");
			s3 = s.pullToken(":");
			s4 = s.pullToken(":");
			config.selection.channels.append(SelectionChannel(s1, s2, s3, s4));
		}
		else {
			usage();
			return 1;
		}
	}
	
	if((argc - optind) != 0){
		usage();
		return 1;
	}

	// Open BDS access
	err = bds.connectService(BString("//") + hostName + "/bdsDataAccess");

	// Connect to service
	if(!err)
		err = bds.connect(user, password);

	if(!err){
		err = testValidateData(bds, config);
	}

	if(err)
		cerr << "Error: " << err.getString() << "\n";

	return err.getErrorNo();
}
